/** *****************************************************
 * Copyright (C) HRider in motion S.L.
 *
 * This file is part of HRider.
 *
 * HRider can not be copied and/or distributed without the express
 * permission of HRider in motion S.L.
 ****************************************************** */
package net.hrider.api.client;

import java.io.InputStream;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.Invocation.Builder;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import net.hrider.api.ApiError;
import net.hrider.api.model.Link;
import net.hrider.api.model.Root;
import org.glassfish.jersey.media.multipart.FormDataMultiPart;
import org.glassfish.jersey.media.multipart.MultiPart;
import org.glassfish.jersey.media.multipart.file.StreamDataBodyPart;

/**
 *
 * @author drh (dani@hrider.net)
 */
public class HRiderService {
    
    public static final String VERSION_PATH = "v1";
    
    public static final String HAL_CONTENT_TYPE = "application/hal+json";    
    public static final String TEXT_URI_LIST_CONTENT_TYPE = "text/uri-list";    
        
    private final ApiClient client;
    private Root rootResource;
    
    private static final Logger LOG = Logger.getLogger(HRiderService.class.getName());
    
    public HRiderService(ApiClient client) {
        this.client = client;
    }
    
    public Root getRootResource() throws ApiException {
        if (rootResource == null) {
            rootResource = get(Root.class, new Link(VERSION_PATH));
        }
        return rootResource;
    }
        
    public <T> T get(Class<T> aclass, Link link) throws ApiException {
        LOG.log(Level.FINE, "get ... {0}", link);
        return get(aclass, link, MediaType.APPLICATION_JSON);
    }
    
    public <T> T get(Class<T> aclass, Link link, String mediaType) throws ApiException {
        LOG.log(Level.FINE, "get ... {0}, {1}, {2}", new Object[] { aclass, link, mediaType });
        Response r = build(mediaType, link).get();
        return readEntity(r, Response.Status.OK, aclass);
    }
    
    public void delete(Link link) throws ApiException {
        LOG.log(Level.FINE, "delete ... {0}", link);
        Response r =  build(link).delete();
        readResponse(r);
    }
    
    public void delete(Link link, String uriList) throws ApiException {
        LOG.log(Level.FINE, "delete ... {0}", link);
        Response r =  build(link).method("DELETE", Entity.entity(uriList, TEXT_URI_LIST_CONTENT_TYPE));
        readResponse(r);
    }
    
    public <T> T create(Link link, T payload) throws ApiException {
        LOG.log(Level.FINE, "create ... {0}, {1}", new Object[] { link, payload });
        Response r = build(link).post(Entity.entity(payload, MediaType.APPLICATION_JSON));        
        return readEntity(r, Response.Status.CREATED, (Class<T>)payload.getClass());                
    }
    
    public <T> T put(Link link, T payload) throws ApiException {
        LOG.log(Level.FINE, "put ... {0}, {1}", new Object[] { link, payload });
        Response r =  build(link).put(Entity.entity(payload, MediaType.APPLICATION_JSON));        
        return readEntity(r, Response.Status.OK, (Class<T>)payload.getClass());
    }
    
    public void put(Link link, String uriList) throws ApiException {
        LOG.log(Level.FINE, "put ... {0}, {1}", new Object[] { link, uriList });
        Response r =  build(link).put(Entity.entity(uriList, TEXT_URI_LIST_CONTENT_TYPE));        
        readResponse(r);
    }
    
    public <T> T post(Link link, T payload) throws ApiException {
        LOG.log(Level.FINE, "post ... {0}, {1}", new Object[] { link, payload });
        Response r =  build(link).post(Entity.entity(payload, MediaType.APPLICATION_JSON));        
        return readEntity(r, Response.Status.OK, (Class<T>)payload.getClass());        
    }
    
    public void uploadData(Link link, String filename, InputStream in) throws ApiException {
        LOG.log(Level.FINE, "uploadData ... {0}, {1}", new Object[] { link });
        StreamDataBodyPart streamDataBodyPart = new StreamDataBodyPart("data", in, filename);        
        MultiPart multipart = new FormDataMultiPart().bodyPart(streamDataBodyPart);
        multipart.setMediaType(MediaType.MULTIPART_FORM_DATA_TYPE);
        
        Response r = getClient().request(MediaType.APPLICATION_JSON, 
                link.getPath()).post(Entity.entity(multipart, multipart.getMediaType()));
        
        readResponse(r);        
    }
    
    public String createUriList(Link[] links) {
        StringBuilder sb = new StringBuilder();
        if (links != null && links.length > 0) {
            for (Link link : links) {
                sb.append(link.getHref()).append("\n");
            }
        }
        return sb.toString();
    }
    
    public String createUriList(long[] ids) {
        StringBuilder sb = new StringBuilder();
        if (ids != null && ids.length > 0) {
            for (long id : ids) {
                sb.append(id).append("\n");
            }
        }
        return sb.toString();
    }
    
    // protected methods *******************************************************
    
    protected ApiClient getClient() {
        return client;
    }
    
    protected <T> T readEntity(Response r, Response.Status statusExpected, Class<T> aclass) throws ApiException {
        int status = r.getStatus();
        if (status == statusExpected.getStatusCode()) {
            return r.readEntity(aclass);
        } else {
            ApiError error;
            try {
                error = r.readEntity(ApiError.class);
            } catch (Exception ex) {
                error = new ApiError(status, ApiError.Code.E999, 
                        String.format("Expected %s (%s) but was %s", statusExpected, 
                                statusExpected.getStatusCode(), status));
            }
            throw new ApiException(Response.Status.fromStatusCode(r.getStatus()), error);            
        }
    }
    
    protected void readResponse(Response r) throws ApiException {
        LOG.log(Level.FINE, "response status: {0}, mediatype: {1}", 
                new Object[] { r.getStatus(), r.getMediaType() });
        
        switch (r.getStatus()) {
            case 200: // ok
            case 204: // no content
                break;
            default:
                ApiError error = r.readEntity(ApiError.class);                
                throw new ApiException(Response.Status.fromStatusCode(r.getStatus()), error);                
        } 
    }
    
    protected Builder build(Link link) {
        return build(MediaType.APPLICATION_JSON, link);
    }
        
    protected Builder build(String mediaType, Link link) {
        return getClient().request(mediaType, link.getQueryParams(), link.getPath());
    }
        
}
